<?php

namespace Ems\Services;

use Util\NovoCodigo;
use Util\Environment;
use Ems\Util\{ ExpectedException, Utilities };
use Ems\Repository\DocumentoRepository;
use Ems\Services\VinculoService;
use Ems\Services\EmailService;
use Doctrine\ORM\EntityManager;
use Util\EmailMessages;

class DocumentoService {
    /** @var Environment */
    protected $environment;

    /** @var DocumentoRepository */
    protected $documentoRepository;

    /** @var Utilities */
    protected $utilities;

    /** @var VinculoService */
    protected $vinculoService;

    /** @var EmailService */
    protected $emailService;

    /** @var OperadorService */
    protected $operadorService;

    /** @var EntityManager */
    protected $entityManager;

    public const CATEG_QR_CODE = 'QRCode';
    public const TIPO_FOTO = 'Foto';
    public const DEFAULT_DOCUMENTO_PUBLICO = 'S';

    public function __construct(
        Environment         $environment,
        DocumentoRepository $documentoRepository,
        Utilities           $utilities,
        VinculoService      $vinculoService,
        OperadorService     $operadorService,
        EmailService        $emailService,
        EntityManager       $entityManager
    ) {
        $this->environment          = $environment;
        $this->documentoRepository  = $documentoRepository;
        $this->utilities            = $utilities;
        $this->vinculoService       = $vinculoService;
        $this->operadorService       = $operadorService;
        $this->emailService         = $emailService;
        $this->entityManager        = $entityManager;
    }

    public function getDocumentosVinculo(
        int $nrVinculo,
        $nrCategoria = null,
        ?string $nomeCateg = null,
        ?string $nomeTipo = null
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRVINCULO'   => $nrVinculo,
            'P_NRCATEGORIA' => $nrCategoria,
            'P_NOMECATEG'   => $nomeCateg,
            'P_NOMETIPO'    => $nomeTipo
        ];

        $documentos = $this->documentoRepository->getDocumentosVinculo(
            $params
        );

        return $documentos;
    }

    public function getCategoriasComDocumentoVinculo(
        int $nrVinculo
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'     => $nrOrg,
            'P_NRVINCULO' => $nrVinculo,
        ];

        $categorias = $this->documentoRepository->getCategoriasComDocumentoVinculo(
            $params
        );

        return $categorias;
    }

    public function getCategorias(): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'     => $nrOrg,
        ];

        $categorias = $this->documentoRepository->getCategorias(
            $params
        );

        return $categorias;
    }

    public function getCategEspecifica(string $nomeCateg): ?array {
        $categoria = $this->documentoRepository->getCategEspecifica($nomeCateg);
        if (empty($categoria)) {
            $this->documentoRepository->criaCategEspecifica($nomeCateg);
            return $this->documentoRepository->getCategEspecifica($nomeCateg);
        }
        return $categoria;
    }

    public function getTipoEspecifico(string $nomeTipo): ?array {
        $tipo = $this->documentoRepository->getTipoEspecifico($nomeTipo);
        if (empty($tipo)) {
            $this->documentoRepository->criaTipoEspecifico($nomeTipo);
            return $this->documentoRepository->getTipoEspecifico($nomeTipo);
        }
        return $tipo;
    }

    public function getTipos(): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'     => $nrOrg,
        ];

        $tipos = $this->documentoRepository->getTipos(
            $params
        );

        return $tipos;
    }

    public function downloadDocumento(
        int $nrAnexo
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $documento = $this->documentoRepository->get(
            $nrOrg, $nrAnexo
        );

        if (!$documento) {
            $message = 'O documento informado não foi encontrado!';
            throw new ExpectedException($message);
        }

        $path = (string) $documento['DSPATH'];
        $file = $this->utilities->downloadFile($path);

        return $file;
    }

    public function saveDocumento(
        array $row,
        $sendEmail = true
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();
        $nrParcNegocio = null;

        $nrCategoria = (int) $row['NRCATEGDOCUMENTO'];
        $nrTipo = (int) $row['NRTIPODOCUMENTO'];

        if (isset($row['NRPARCNEGOCIO'])) {
            $nrParcNegocio = $row['NRPARCNEGOCIO'];
        } else {
            $nrVinculo = (int) $row['NRVINCULOM'];

            $nrParcNegocio = $this->vinculoService->getNrParcNegocioVinculo(
                $nrVinculo
            );
        }

        // Salva o arquivo no File Server
        $filePath = $this->saveFileOnServer($row);

        $nrAnexo = NovoCodigo::geraCodigo(
            $nrOrg, 'GPE_ANEXOS', 1, 10, true
        );

        $base64 = (string) $row['FILE'];
        $type = $this->utilities->getFileMIMEType($base64);

        $params = [
            'DSPATH'        => $filePath,
            'DSREFERENCIA'  => $row['DSREFERENCIA'],
            'IDFOTO'        => 'N',
            'IDTYPE'        => $type,
            'NMANEXO'       => $row['NMANEXO'],
            'NRANEXO'       => $nrAnexo,
            'NRCATEGORIA'   => $nrCategoria,
            'NRORG'         => $nrOrg,
            'NRPARC'        => $nrParcNegocio,
            'NRTIPO'        => $nrTipo,
        ];

        // if ($sendEmail) {
        //     $this->entityManager->beginTransaction();
        // }

        try {
            $this->documentoRepository->saveDocumento($params);
        } catch (\Exception $e) {
            $this->utilities->deleteFile($filePath);
            throw $e;
        }

        if ($sendEmail) {
            $file = [
                'base64'    => $base64,
                'name'      => $row['NMANEXO'],
            ];

            $row['FILE'] = $file;

            try {
                $this->sendEmailNovoDocumento($row);
            } catch (\Exception $e) {
                // $this->entityManager->rollback();
                // throw $e;
            }

            // $this->entityManager->commit();
        }
    }

    public function saveDocumentos(
        array $row,
        ?string $nomeCateg = null,
        ?string $nomeTipo = null,
        ?bool $updateExistent = true
    ) {
        $arrayDocs = (array) ($row['DOCUMENTOS'] ?? []);
        $nrOrg = (int) $this->environment->getNrOrg();
        
        foreach ($arrayDocs as $doc) {
            $nomeCategoria = $doc['NMCATEGDOCUMENTO'] ?? $nomeCateg ?? null;
            $nomeTipo = $doc['NMTIPODOCUMENTO'] ?? $nomeTipo ?? null;
            $nrVinculom = (int) $doc['NRVINCULOM'];
            $nrParcNegocio = $this->vinculoService->getNrParcNegocioVinculo($nrVinculom);

            // Salva o arquivo no File Server
            $filePath = $this->saveFileOnServer($doc);

            $categoria = $this->getCategEspecifica($nomeCategoria);
            $nrCategoria = $categoria['NRCATEGDOCUMENTO'] ?? null;

            $tipo = $this->getTipoEspecifico($nomeTipo);
            $nrTipo = $tipo['NRTIPODOCUMENTO'] ?? null;
    
            $base64 = (string) $doc['FILE'];
            $type = $this->utilities->getFileMIMEType($base64);

            /* Atualizacao anexo */

            $existentDocs = $this->getDocumentosVinculo($nrVinculom, null, $nomeCategoria, $nomeTipo);
            $existentDoc = reset($existentDocs);

            /* Cria se for informado ou se um documento ainda nao existir */

            if (!$updateExistent || empty($existentDoc)) {
                $nrAnexo = NovoCodigo::geraCodigo(
                    $nrOrg, 'GPE_ANEXOS', 1, 10, true
                );
                
                $params = [
                    'DSPATH'        => $filePath,
                    'DSREFERENCIA'  => $doc['DSREFERENCIA'],
                    'IDFOTO'        => 'N',
                    'IDTYPE'        => $type,
                    'NMANEXO'       => $doc['NMANEXO'],
                    'NRANEXO'       => $nrAnexo,
                    'NRCATEGORIA'   => $nrCategoria,
                    'NRORG'         => $nrOrg,
                    'NRPARC'        => $nrParcNegocio,
                    'NRTIPO'        => $nrTipo,
                ];
                
                $this->documentoRepository->saveDocumento($params);
                continue;
            }

            /* Atualizacao documento existente */

            if (empty($existentDoc)) {
                continue;
            }

            if (!empty($existentDoc['DSPATH'])) {
                $this->utilities->deleteFile($existentDoc['DSPATH']);
            }

            $this->documentoRepository->updateFile($existentDoc['NRANEXO'], $filePath);
        }
    }

    private function sendEmailNovoDocumento(
        array $row
    ) {
        $cdOperador = $row['CDOPERADOR'] ?? '1';
        $nrVinculo = (int) $row['NRVINCULOM'];
        $file = (array) $row['FILE'];

        $operador = $this->operadorService->get($cdOperador);
        $vinculo = $this->vinculoService->getDadosVinculo($nrVinculo);
        $operadoresDP = $this->operadorService->getOperadoresDp();

        $nmOperador = (string) $operador['NMOPERADOR'];
        $nmVinculo = (string) $vinculo['NMVINCULOM'];
        $nrEstrutLegal = $vinculo['NRESTRUTLEGAL'];

        if ($nrEstrutLegal === null) {
            $message = "O vínculo $nmVinculo não possui estrutura legal!";
            throw new ExpectedException($message);
        }

        $translationSubject = [
            '@VINCULO@' => $nmVinculo,
        ];

        $subjectTemplate = EmailMessages::DOCUMENTO_SUBJECT;
        $subject = strtr($subjectTemplate, $translationSubject);

        $content = $this->getEmailContent(
            $nmOperador, $nmVinculo, $row
        );

        $attachment = Utilities::blobToFile(
            $file['base64'], $file['name']
        );

        $this->emailService->enviarEmailsOcorrenciaDocumento(
            $operadoresDP, $subject, $content,
            $nrEstrutLegal, $attachment
        );
    }

    private function getEmailContent(
        string $nmOperador,
        string $nmVinculo,
        array $row
    ): string {
        $referencia = (string) $row['DSREFERENCIA'];
        $nrCategoria = (int) $row['NRCATEGDOCUMENTO'];
        $nrTipo = (int) $row['NRTIPODOCUMENTO'];

        $categoria = $this->getDadosCategoria($nrCategoria);
        $nmCategoria = (string) $categoria['NMCATEGDOCUMENTO'];

        $tipo = $this->getDadosTipo($nrTipo);
        $nmTipo = (string) $tipo['NMTIPODOCUMENTO'];

        $horario = date('d/m/Y H:i');

        $translations = [
            '@OPERADOR@' => $nmOperador,
            '@VINCULO@' => $nmVinculo,
            '@REFERENCIA@' => $referencia,
            '@CATEGORIA@' => $nmCategoria,
            '@TIPO@' => $nmTipo,
            '@HORARIO@' => $horario,
        ];

        $baseMessage = EmailMessages::DOCUMENTO_BODYMESSAGE;

        $content = strtr($baseMessage, $translations);

        return $content;
    }

    public function saveFileOnServer(
        array $row
    ): string {
        $nrOrg = (int) $this->environment->getNrOrg();

        $base64 = (string) $row['FILE'];

        $directory = "EMS/$nrOrg/Documentos/";
        $name = $row['NMANEXO'] ?? $row['NMFILE'];

        $filePath = $this->utilities->saveFileOnFileServer(
            $base64, null, $nrOrg, $directory,
            $name, null, false
        );

        return $filePath;
    }

    public function getDadosCategoria(
        int $nrCategoria
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();

        $categoria = $this->documentoRepository->getDadosCategoria(
            $nrOrg, $nrCategoria
        );

        return $categoria;
    }

    public function getDadosTipo(
        int $nrtipo
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();

        $tipo = $this->documentoRepository->getDadosTipo(
            $nrOrg, $nrtipo
        );

        return $tipo;
    }
}